use std::{fmt::Display, path::PathBuf};

use crate::run::{BENCH_BASE_PATH, BENCH_DATA_PATH, BENCH_BIN_PATH, BENCH_LOG_PATH};
use crate::run::args::TestType::*;
use crate::run::args::FilterType::*;

const CHUNK_MAX_BITS: u64 = 18;
const ADDRESS_MIN_BITS: u64 = 3;


#[derive(Debug, Copy, Clone)]
pub struct BenchArgs {
    pub n: u64,
    pub data_args: DataArgs,
    pub bloom_filter_args: FilterArgs,
    pub scan_args: ScanArgs
}

impl Display for BenchArgs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "BenchArgs({}, {}, {}, {})", self.n, self.data_args, self.bloom_filter_args, self.scan_args)
    }
}

#[derive(Debug, Copy, Clone)]
pub struct DataArgs {
    pub seed: u64,
    pub scan_subnet_size: u64,
    pub data_set_index: u64,
}

impl Display for DataArgs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Data(seed: {}, subnet: {}, dataset index: {})", self.seed, self.scan_subnet_size, self.data_set_index)
    }
}

#[derive(Debug, Copy, Clone)]
pub struct FilterArgs {
    pub test_type: TestType,
    pub filter_type: FilterType,
    pub address_bits: u64,
    pub chunk_address_bits: u64,
    pub hash_count: Option<u64>,
}

impl Display for FilterArgs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "BloomFilter(type: {}, address bits: {}, hashes: {:?})",self.test_type, self.address_bits, self.hash_count)
    }
}

#[derive(Debug, Copy, Clone)]
pub struct ScanArgs {
    pub rate: u64
}

impl Display for ScanArgs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Scan(rate: {})", self.rate)
    }
}

impl BenchArgs {
    pub fn rel_path(&self) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(PathBuf::from(self.n.to_string()));
        path.push(self.data_args.rel_path());
        path.push(self.bloom_filter_args.rel_path());
        path.push(self.scan_args.rel_path());
        return path;
    }

    pub fn wd_path(&self) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_DATA_PATH);
        path.push(self.rel_path());
        return path;
    }

    pub fn bin_path(&self) -> PathBuf {
        Self::bin_bin_path(self.data_args, self.bloom_filter_args)
    }

    pub fn bin_bin_path(data_args: DataArgs, bloom_args: FilterArgs) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_BIN_PATH);
        path.push(data_args.rel_path());
        path.push(bloom_args.rel_path());
        return path;
    }

    pub fn bin_log_path(data_args: DataArgs, bloom_args: FilterArgs) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_LOG_PATH);
        path.push("build");
        path.push(data_args.rel_path());
        path.push(bloom_args.rel_path());
        return path;
    }

    pub fn bin_wd_path(data_args: DataArgs, bloom_args: FilterArgs) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_DATA_PATH);
        path.push("build");
        path.push(data_args.rel_path());
        path.push(bloom_args.rel_path());
        return path;
    }

    pub fn log_path(&self) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(BENCH_BASE_PATH);
        path.push(BENCH_LOG_PATH);
        path.push(self.rel_path());
        return path;
    }
}

impl DataArgs {
    pub fn from(seed: u64, scan_subnet_size: u64, data_set_index: u64) -> Self {
        Self {
            seed,
            scan_subnet_size,
            data_set_index,
        }
    }

    pub fn rel_path(&self) -> PathBuf {
        let mut path = PathBuf::new();
        path.push(self.seed.to_string());
        path.push(self.scan_subnet_size.to_string());
        path.push(self.data_set_index.to_string());
        return path;
    }
}

impl FilterArgs {
    pub fn from(data_args: DataArgs, test_type: TestType, filter_type: FilterType, entries: u64) -> FilterArgs {
        match test_type {
            Normal(false_hit_rate) | BpfStats(false_hit_rate) => {
                match filter_type {
                    Bloom => {
                        let hash_count = (-false_hit_rate.log2()).round();
                        let size = (entries as f64) * (hash_count/ 2f64.ln());
                        let address_bits = size.log2().round() as u64;
                        let address_bits = ADDRESS_MIN_BITS.max(address_bits);
                        let chunk_address_bits = CHUNK_MAX_BITS.min(address_bits);

                        FilterArgs {
                            test_type,
                            filter_type,
                            address_bits,
                            chunk_address_bits,
                            hash_count: Some(hash_count as u64)
                        }
                    },
                    Bitmap => {
                        let address_bits = ADDRESS_MIN_BITS.max(data_args.scan_subnet_size);
                        let chunk_address_bits = CHUNK_MAX_BITS.min(address_bits);

                        FilterArgs {
                            test_type,
                            filter_type,
                            address_bits,
                            chunk_address_bits,
                            hash_count: None
                        }
                    }
                }
            }
            Baseline => FilterArgs {
                test_type,
                filter_type,
                address_bits: ADDRESS_MIN_BITS,
                chunk_address_bits: ADDRESS_MIN_BITS,
                hash_count: None
            },
        }
    }

    pub fn rel_path(&self) -> PathBuf {
        if let Baseline = self.test_type {
            return PathBuf::from("baseline")
        } else {
            let mut name = self.address_bits.to_string();
            match self.filter_type {
                Bloom => {
                    name.push_str("-bloom-");
                    name.push_str(self.hash_count.unwrap().to_string().as_str());
                },
                Bitmap => name.push_str("-bitmap"),
            }
            if let BpfStats(_) = self.test_type {
                name.push_str("-bpf")
            }
            return PathBuf::from(name);
        }
    }
}

impl ScanArgs {
    pub fn new(rate: u64) -> Self {
        Self {
            rate
        }
    }

    pub fn rel_path(&self) -> PathBuf {
        PathBuf::from(self.rate.to_string())
    }
}

#[derive(Debug, Copy, Clone)]
pub enum TestType {
    Baseline, BpfStats(f64), Normal(f64)
}

impl Display for TestType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}

#[derive(Debug, Copy, Clone)]
pub enum FilterType {
    Bloom, Bitmap
}

impl Display for FilterType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}
