use std::{path::PathBuf, time::Instant, net::Ipv4Addr, fs::File, io::{Write, BufReader, BufRead}};

use anyhow::anyhow;
use clap::Parser;
use log::info;
use responder_common::filter;
use responder_common::filter::bloom_filter;

#[derive(Debug, Parser)]
struct Options {
    pub filter_type: String,
    pub ip_path: PathBuf,
    pub out_path: Option<PathBuf>,
    #[clap(short, long)]
    pub force: bool,
    #[clap(short, long)]
    pub timing_path: Option<PathBuf>
}

fn main() -> Result<(), anyhow::Error> {
    env_logger::init();
    let opts = Options::parse();

    return build_filter(opts);
}

fn build_filter(opts: Options) -> Result<(), anyhow::Error> {
    let ext_suff = match opts.filter_type.as_str() {
        "bloom" => ".bfb",
        "bitmap" => ".fb",
        _ => return Err(anyhow!("Expected filter type 'bloom' or 'bitmap', got {}", opts.filter_type))
    };

    let out_path = opts.out_path.unwrap_or_else(|| {
        let mut ext = opts.ip_path.extension().unwrap_or_default().to_os_string();
        ext.push(ext_suff);
        opts.ip_path.with_extension(ext)
    });

    if out_path.exists() {
        if !opts.force {
            return Err(anyhow!("Output path {} already exists, use --force/-f to overwrite", out_path.display()));
        } else if !out_path.is_file() {
            return Err(anyhow!("Can not force overwrite output path {}: is not a file", out_path.display()));
        }
    }

    let start = Instant::now();


    let filter = start.elapsed();

    let mut file = File::create(out_path)?;
    file.write(&(filter::ADDRESS_BITS as u64).to_be_bytes())?;
    file.write(&(filter::ADDRESS_BITS_CHUNK as u64).to_be_bytes())?;

    let filter_vec = match opts.filter_type.as_str() {
        "bloom" => {
            file.write(&(bloom_filter::HASH_COUNT as u64).to_be_bytes())?;
            build_bloom_filter(opts.ip_path)
        }
        "bitmap" => build_bitmap_filter(opts.ip_path),
        _ => Err(anyhow!("Expected 'bloom' or 'bitmap' filtertype, got {}", opts.filter_type))
    }?;

    let byte_slice: &[u8] = filter_vec.as_slice();
    file.write_all(byte_slice)?;

    let end = start.elapsed();

    info!("build filter: {:?} write: {:?} | total: {:?}", filter, end-filter, end);

    if let Some(path) = opts.timing_path {
        File::create(path)?.write_all(format!("{{\"write\": {}, \"build\": {}}}\n",(end-filter).as_nanos(), filter.as_nanos()).as_bytes())?;
    }

    Ok(())
}

fn build_bloom_filter(ip_path: PathBuf) -> anyhow::Result<Vec<u8>> {
    // let mut addresses = Vec::new();
    let mut bloom_filter_vec: Vec<u8> = Vec::with_capacity(filter::BITS>>3);
    bloom_filter_vec.resize(filter::BITS >> 3, 0);

    let f = File::open(ip_path)?;
    let reader = BufReader::new(f);
    let mut clash_count = 0;
    for (i, line) in reader.lines().enumerate() {
        if i % 1_000_000 == 0{
            info!("processed {}M ips; {} clashes",i / 1_000_000, clash_count);
        }
        let line = line?;
        let key = u32::from(line.parse::<Ipv4Addr>()?);
        // addresses.push(key);
        let mut clashes = true;
        info!("{}", bloom_filter::HASH_COUNT);
        for i in 0..bloom_filter::HASH_COUNT {
            let hash = bloom_filter::hash(key, i);
            let i = hash >> 3;
            let byte_i = hash & 7;
            // info!("{} {} {} {}", line, hash, i, byte_i);
            clashes = clashes && ((bloom_filter_vec[i as usize] >> (7 - byte_i)) & 1 == 1);
            bloom_filter_vec[i as usize] |= 0x80 >> byte_i;
        }
        if clashes {
            clash_count += 1;
        }
    }
    info!("clash count: {}", clash_count);
    return Ok(bloom_filter_vec);
}

fn build_bitmap_filter(ip_path: PathBuf) -> anyhow::Result<Vec<u8>> {
    let mut filter_vec: Vec<u8> = Vec::with_capacity(filter::BITS>>3);
    filter_vec.resize(filter::BITS >> 3, 0);

    let f = File::open(ip_path)?;
    let reader = BufReader::new(f);

    for (i, line) in reader.lines().enumerate() {
        if i % 1_000_000 == 0{
            info!("processed {}M ips",i / 1_000_000);
        }
        let line = line?;
        let key = u32::from(line.parse::<Ipv4Addr>()?);
        let key = key & (filter::ADDRESS_MASK as u32);
        // addresses.push(key);
        let i = key >> 3;
        let byte_i = key & 7;
        // info!("{} {} {} {}", line, hash, i, byte_i);
        filter_vec[i as usize] |= 0x80 >> byte_i;
    }

    return Ok(filter_vec);
}
